# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/iiif.ipynb.

# %% auto 0
__all__ = ['IIIF_CONTEXT', 'JPEG_FORMAT', 'IIIF', 'Svg']

# %% ../nbs/iiif.ipynb 4
import numpy as np
from xml.dom.minidom import parseString
from svg.path import parse_path
import cv2
import numpy as np
from svg.path import parse_path
from xml.dom.minidom import parseString
import svgwrite
import json
import glob
import os
from PIL import Image
from bs4 import BeautifulSoup, Tag, NavigableString
import datetime
from xsd_validator import XsdValidator
import requests
import xml.etree.ElementTree as ET
from typing import List, Dict, Union

# %% ../nbs/iiif.ipynb 5
IIIF_CONTEXT = "http://iiif.io/api/presentation/3/context.json"
JPEG_FORMAT = "image/jpeg"

class IIIF:

    """
    A class representing the International Image Interoperability Framework (IIIF) for images.
    """

    @staticmethod
    def create_iiif_manifest_from_local_dir(root_path: str, target_path: str, output_path: str, item_id: str) -> None:
        """
        Create an IIIF manifest from a local directory containing .jpg images.

        Args:
            root_path (str): The root path of the images.
            target_path (str): The target path containing the images.
            output_path (str): The path to save the manifest file.
            item_id (str): The ID for the IIIF item.

        Returns:
            None
        """
        image_files = IIIF._retrieve_image_files(target_path)
        canvases = IIIF._generate_canvases(image_files, root_path, item_id)
        manifest = IIIF._generate_manifest(canvases, item_id)
        IIIF._write_manifest_to_file(manifest, output_path)

    @staticmethod
    def _retrieve_image_files(target_path: str) -> List[str]:
        """Retrieve image files from the target path.

        Args:
            target_path (str): The path containing the images.

        Returns:
            list: Sorted list of image file paths.
        """
        return sorted(glob.glob(os.path.join(target_path, "**/*.jpg")))

    @staticmethod
    def _generate_canvases(image_files: List[str], root_path: str, item_id: str) -> List[Dict[str, any]]:
        """
        Generate IIIF canvases from image files.

        Args:
            image_files (list): List of image file paths.
            root_path (str): The root path of the images.
            item_id (str): The ID for the IIIF item.

        Returns:
            list: List of IIIF canvases.
        """
        canvases = []
        prefix_url = f"/iiif/{item_id}"

        for idx, file_path in enumerate(image_files):
            with Image.open(file_path) as img:
                width, height = img.size

            count = idx + 1
            canvas = {
                "id": f"{prefix_url}/canvas/p{count}",
                "type": "Canvas",
                "height": height,
                "width": width,
                "items": [
                    {
                        "type": "AnnotationPage",
                        "items": [
                            {
                                "id": f"{prefix_url}/annotation/p{str(count).zfill(4)}-image",
                                "type": "Annotation",
                                "motivation": "painting",
                                "body": {
                                    "id": file_path.replace(root_path, ""),
                                    "type": "Image",
                                    "format": JPEG_FORMAT,
                                    "height": height,
                                    "width": width
                                },
                                "target": f"{prefix_url}/canvas/p{count}"
                            }
                        ]
                    }
                ]
            }
            canvases.append(canvas)

        return canvases

    @staticmethod
    def _generate_manifest(canvases: List[Dict[str, any]], item_id: str) -> Dict[str, any]:
        """
        Generate IIIF manifest.

        Args:
            canvases (list): List of IIIF canvases.
            item_id (str): The ID for the IIIF item.

        Returns:
            dict: IIIF manifest dictionary.
        """
        return {
            "@context": IIIF_CONTEXT,
            "id": f"/iiif/{item_id}/manifest.json",
            "type": "Manifest",
            "label": item_id,
            "items": canvases
        }

    @staticmethod
    def _write_manifest_to_file(manifest: Dict[str, any], output_path: str) -> None:
        """
        Write the IIIF manifest to a file.

        Args:
            manifest (dict): IIIF manifest dictionary.
            output_path (str): The path to save the manifest file.
        """
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, "w") as file:
            json.dump(manifest, file, ensure_ascii=False, indent=4)

    @staticmethod
    def points_to_svg_path(points: str) -> str:
        """
        Convert a list of points into an SVG path.

        Args:
            points: String of comma-separated coordinates.

        Returns:
            SVG path representation.
        """
        coordinates = [int(i) for i in points.split(",")]
        path = f"M {coordinates[0]} {coordinates[1]}"
        for i in range(2, len(coordinates), 2):
            path += f" L {coordinates[i]} {coordinates[i+1]}"
        path += " Z"
        return path

    @staticmethod
    def convert_ndl_ocr_xml_to_iiif_manifest(xml_path: str, base_manifest_path: str, output_manifest_path: str) -> None:
        """
        Convert NDL OCR XML to IIIF manifest.

        Args:
            xml_path: Path to the NDL OCR XML file.
            base_manifest_path: Path to the base IIIF manifest.
            output_manifest_path: Path to save the converted IIIF manifest.

        """
        with open(base_manifest_path) as f:
            manifest_base = json.load(f)

        canvases = manifest_base["items"]        

        # Load XML content
        with open(xml_path, "r") as file:
            content = file.read()

        # Create a BeautifulSoup object
        soup = BeautifulSoup(content, "lxml")

        xml_pages = soup.find_all("page")

        stored_annotations: List[List[Dict[str, Union[str, Dict[str, any]]]]] = []

        for i, xml_page in enumerate(xml_pages):
            canvas_id = canvases[i]["id"]
            annotationPageItems = []

            for xml_block in xml_page.find_all("block"):
                annotationPageItems.append(IIIF._process_block(xml_block, canvas_id, len(annotationPageItems)))

            for xml_line in xml_page.find_all("line"):
                annotationPageItems.append(IIIF._process_line(xml_line, canvas_id, len(annotationPageItems)))

            for xml_polygon in xml_page.find_all("polygon"):
                annotationPageItems.append(IIIF._process_polygon(xml_polygon, canvas_id, len(annotationPageItems)))

            stored_annotations.append(annotationPageItems)


        for i, canvas in enumerate(canvases):

            annotationPageItems = stored_annotations[i]

            canvas["annotations"] = [
                {
                    "type": "AnnotationPage",
                    "items": annotationPageItems
                }
            ]

        with open(output_manifest_path, "w") as f:
            json.dump(manifest_base, f, indent=4, ensure_ascii=False)

    @staticmethod
    def _process_block(xml_block, canvas_id: str, index: int) -> Dict[str, any]:
        """
        Process XML block and return the corresponding annotation.

        Args:
            xml_block: The block element from the XML.
            canvas_id: The canvas ID for the current page.

        Returns:
            An annotation dictionary for the given block.
        """
        # TODO: Implement this using the block processing logic from above
        x = xml_block["x"]
        y = xml_block["y"]
        w = xml_block["width"]
        h = xml_block["height"]

        string = xml_block["string"] if "string" in xml_block.attrs else ""

        return {
                "id": f"a_{index}_{x}_{y}_{w}_{h}",
                "type": "Annotation",
                "_motivation": "tagging",
                "motivation": "commenting",
                "body": {
                    "type": "TextualBody",
                    "_value": xml_block["type"],
                    "__value": xml_block["type"],
                    "value": f'⭐️【{xml_block["type"]}】{string}'

                },
                "target": f"{canvas_id}#xywh={x},{y},{w},{h}"
            }
        

    @staticmethod
    def _process_line(xml_line, canvas_id: str, index: int) -> Dict[str, any]:
        """
        Process XML line and return the corresponding annotation.

        Args:
            xml_line: The line element from the XML.
            canvas_id: The canvas ID for the current page.

        Returns:
            An annotation dictionary for the given line.
        """
        # TODO: Implement this using the line processing logic from above
        x = xml_line["x"]
        y = xml_line["y"]
        w = xml_line["width"]
        h = xml_line["height"]

        order = xml_line["order"] if "order" in xml_line.attrs else "-1"

        string = xml_line["string"] if "string" in xml_line.attrs else ""

        return {
                "id": f"a_{index}_{x}_{y}_{w}_{h}",
                "type": "Annotation",
                "_motivation": "tagging",
                "motivation": "commenting",
                "body": {
                    "type": "TextualBody",
                    "value": f'【{xml_line["type"]}:{order}】{string}'

                },
                "target": f"{canvas_id}#xywh={x},{y},{w},{h}"
            }

    @staticmethod
    def _process_polygon(xml_polygon, canvas_id: str, index: int) -> Dict[str, any]:
        """
        Process XML polygon and return the corresponding annotation.

        Args:
            xml_polygon: The polygon element from the XML.
            canvas_id: The canvas ID for the current page.

        Returns:
            An annotation dictionary for the given polygon.
        """
        # TODO: Implement this using the polygon processing logic from above
        points = xml_polygon["points"] # .split(",")
        p = IIIF.points_to_svg_path(points)

        return {
                "id": f"a_{index}_{points}",
                "type": "Annotation",
                "motivation": "tagging",
                "body": {
                    "type": "TextualBody",
                    "value": "TEXTBLOCK"

                },
                "target": {
                            "type": "SpecificResource",
                            "source": canvas_id,
                            "selector": {
                                "type": "SvgSelector",
                                "value": f"<svg xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'><g><path d='{p}' stroke='red' fill='none'/></g></svg>"
                            }
                        }
            }

    @staticmethod
    def _create_and_set_tag(soup, tag_name, string_content=None, parent=None):
        """
        Helper function to create a tag, set its content, and optionally attach it to a parent tag.

        Args:
            soup: The BeautifulSoup object used for creating the tag.
            tag_name: The name of the tag to be created.
            string_content: The content to be set inside the tag.
            parent: The tag to which the created tag should be attached.

        Returns:
            The created tag.
        """
        tag = soup.new_tag(tag_name)
        if string_content:
            tag.string = string_content
        if parent:
            parent.append(tag)
        return tag

    @staticmethod
    def set_desc(alto: BeautifulSoup, soup_alto: BeautifulSoup, opt_str: str, filename: str) -> BeautifulSoup:
        """
        Set description details to an ALTO (Analyzed Layout and Text Object) structure.

        Args:
            alto: A BeautifulSoup object representing the ALTO structure to be modified.
            soup_alto: The BeautifulSoup object used to create new tags.
            opt_str: A string representing OCR processing options.
            filename: The name of the source image file.

        Returns:
            The modified BeautifulSoup object representing the updated ALTO structure.
        """

        '''
        alto_description = soup_alto.new_tag("Description")
        alto.append(alto_description)

        alto_description_measurement_unit = soup_alto.new_tag("MeasurementUnit")
        alto_description_measurement_unit.string = "pixel"
        alto_description.append(alto_description_measurement_unit)

        alto_description_source_image_information = soup_alto.new_tag("sourceImageInformation")
        alto_description.append(alto_description_source_image_information)

        alto_description_source_image_information_file_name = soup_alto.new_tag("fileName")
        alto_description_source_image_information_file_name.string = filename
        alto_description_source_image_information.append(alto_description_source_image_information_file_name)

        alto_description_ocrprocessing = soup_alto.new_tag("OCRProcessing")
        alto_description.append(alto_description_ocrprocessing)
        alto_description_ocrprocessing["ID"] = "OCR_1" 

        alto_description_ocrprocessing_processingStep = soup_alto.new_tag("ocrProcessingStep")
        alto_description_ocrprocessing.append(alto_description_ocrprocessing_processingStep)
        

        alto_description_ocrprocessing_processingStep_processingDateTime = soup_alto.new_tag("processingDateTime")

        dt_now = datetime.datetime.now(datetime.timezone(datetime.timedelta(hours=9)))

        alto_description_ocrprocessing_processingStep_processingDateTime.string = dt_now.isoformat()
        alto_description_ocrprocessing_processingStep.append(alto_description_ocrprocessing_processingStep_processingDateTime)

        alto_description_ocrprocessing_processingStep_processingAgency = soup_alto.new_tag("processingAgency")
        alto_description_ocrprocessing_processingStep.append(alto_description_ocrprocessing_processingStep_processingAgency)
        alto_description_ocrprocessing_processingStep_processingAgency.string = "Google Colaboratory"

        alto_description_ocrprocessing_processingStep_processingStepSettings = soup_alto.new_tag("processingStepSettings")
        alto_description_ocrprocessing_processingStep.append(alto_description_ocrprocessing_processingStep_processingStepSettings)
        alto_description_ocrprocessing_processingStep_processingStepSettings.string = f'{opt_str}'
        
        alto_description_ocrprocessing_processingStep_processingSoftware = soup_alto.new_tag("processingSoftware")
        alto_description_ocrprocessing_processingStep.append(alto_description_ocrprocessing_processingStep_processingSoftware)


        alto_description_ocrprocessing_processingStep_processingSoftware_softwareCreator = soup_alto.new_tag("softwareCreator")
        alto_description_ocrprocessing_processingStep_processingSoftware.append(alto_description_ocrprocessing_processingStep_processingSoftware_softwareCreator)
        alto_description_ocrprocessing_processingStep_processingSoftware_softwareCreator.string = "National Diet Library"

        alto_description_ocrprocessing_processingStep_processingSoftware_softwareName = soup_alto.new_tag("softwareName")
        alto_description_ocrprocessing_processingStep_processingSoftware.append(alto_description_ocrprocessing_processingStep_processingSoftware_softwareName)
        alto_description_ocrprocessing_processingStep_processingSoftware_softwareName.string = "NDLOCR"

        alto_description_ocrprocessing_processingStep_processingSoftware_softwareVersion = soup_alto.new_tag("softwareVersion")
        alto_description_ocrprocessing_processingStep_processingSoftware.append(alto_description_ocrprocessing_processingStep_processingSoftware_softwareVersion)
        alto_description_ocrprocessing_processingStep_processingSoftware_softwareVersion.string = "2.1"
        '''

        desc = IIIF._create_and_set_tag(soup_alto, "Description", parent=alto)
        
        IIIF._create_and_set_tag(soup_alto, "MeasurementUnit", "pixel", desc)

        source_img_info = IIIF._create_and_set_tag(soup_alto, "sourceImageInformation", parent=desc)
        IIIF._create_and_set_tag(soup_alto, "fileName", filename, source_img_info)

        ocr_processing = IIIF._create_and_set_tag(soup_alto, "OCRProcessing", parent=desc)
        ocr_processing["ID"] = "OCR_1"

        ocr_processing_step = IIIF._create_and_set_tag(soup_alto, "ocrProcessingStep", parent=ocr_processing)
        
        dt_now = datetime.datetime.now(datetime.timezone(datetime.timedelta(hours=9)))
        IIIF._create_and_set_tag(soup_alto, "processingDateTime", dt_now.isoformat(), ocr_processing_step)
        IIIF._create_and_set_tag(soup_alto, "processingAgency", "Google Colaboratory", ocr_processing_step)
        IIIF._create_and_set_tag(soup_alto, "processingStepSettings", opt_str, ocr_processing_step)

        processing_software = IIIF._create_and_set_tag(soup_alto, "processingSoftware", parent=ocr_processing_step)

        software_details = {
            "softwareCreator": "National Diet Library",
            "softwareName": "NDLOCR",
            "softwareVersion": "2.1"
        }

        for tag, value in software_details.items():
            IIIF._create_and_set_tag(soup_alto, tag, value, processing_software)

        return alto
    
    @staticmethod
    def _pretty_print(current, parent = None, index: int = -1, depth: int = 0) -> None:
        """
        Recursive function to pretty print XML elements with indentation.

        Args:
            current: The current XML element being processed.
            parent: The parent of the current XML element.
            index: The index of the current XML element in its parent's children list.
            depth: Current depth for indentation.
        """
        for i, node in enumerate(current):
            IIIF._pretty_print(node, current, i, depth + 1)
        if parent is not None:
            if index == 0:
                parent.text = '\n' + ('\t' * depth)
            else:
                parent[index - 1].tail = '\n' + ('\t' * depth)
            if index == len(parent) - 1:
                current.tail = '\n' + ('\t' * (depth - 1))

    @staticmethod
    def format_and_save(html: str, output_path: str) -> None:
        """
        Reformat the provided HTML content and save it as an XML file.

        Args:
            html: HTML content to be reformatted.
            output_path: The file path where the reformatted content should be saved.
        """
        root = ET.fromstring(html)
        ET.register_namespace('', 'http://www.loc.gov/standards/alto/ns-v4#')

        IIIF._pretty_print(root)
        tree = ET.ElementTree(root)
        tree.write(output_path, encoding="utf-8", xml_declaration=True)

    @staticmethod
    def validate(output_path: str) -> None:
        """
        Validate the XML file against the ALTO v4.2 schema.

        Args:
            output_path: The file path of the XML to be validated.
        """
        schema_path = "/tmp/url.xml"
        schema_url = "http://www.loc.gov/standards/alto/v4/alto-4-2.xsd"

        if not os.path.exists(schema_path):

            response = requests.get(schema_url)
            with open(schema_path, 'wb') as file:
                file.write(response.content)
        
        validator = XsdValidator(schema_path)
        validator.assert_valid(output_path)

    @staticmethod
    def convert_ndl_ocr_xml_to_alto_xml(xml_path, output_dir, opt_json_path):
        """
        Convert NDL OCR XML to ALTO XML format.

        Args:
            xml_path: The file path to the source XML.
            output_dir: The directory where the output ALTO XML files should be saved.
            opt_json_path: The file path to the JSON options file.
        """
        item_id = os.path.basename(xml_path).split(".")[0]

        with open(opt_json_path, "r") as f:
            opt_json = json.load(f)
        
        opt_str = json.dumps(opt_json, ensure_ascii=False)

        # Load XML content
        with open(xml_path, "r") as file:
            content = file.read()

        # Create BeautifulSoup object
        soup = BeautifulSoup(content, "lxml")
        xml_pages = soup.find_all("page")        

        for i, xml_page in enumerate(xml_pages):
            soup_alto = BeautifulSoup(features="xml")
            alto = soup_alto.new_tag("alto", 
                                     xmlns="http://www.loc.gov/standards/alto/ns-v4#", 
                                     **{"xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance", 
                                        "xsi:schemaLocation": "http://www.loc.gov/standards/alto/ns-v4# http://www.loc.gov/standards/alto/v4/alto-4-2.xsd"})
            soup_alto.append(alto)

            filename = f"{str(i + 1).zfill(4)}.xml"
            alto = IIIF.set_desc(alto, soup_alto, opt_str, f"{item_id}_{filename}")
            
            layout = soup_alto.new_tag("Layout")
            alto.append(layout)
            page = soup_alto.new_tag("Page", ID="Page_" + str(i + 1), PHYSICAL_IMG_NR=str(i + 1))
            layout.append(page)

            printSpace = soup_alto.new_tag("PrintSpace", ID="PrintSpace_" + str(i + 1))
            page.append(printSpace)
            
            ex_blocks = []
            ex_lines = []

            def createLine(line, ex_lines):
                textLine = soup_alto.new_tag("TextLine")
                textBlock.append(textLine)
                
                # textLine["CONTENT"] = line["string"]

                ex_lines.append(line)
                textLine["ID"] = "TextLine_" + str(len(ex_lines))

                textLine["HPOS"] = line["x"]
                textLine["VPOS"] = line["y"]
                textLine["WIDTH"] = line["width"]
                textLine["HEIGHT"] = line["height"]

                textString = soup_alto.new_tag("String")
                textLine.append(textString)
                textString["ID"] = "Word_" + str(len(ex_lines))
                textString["CONTENT"] = line["string"]

                textString["HPOS"] = line["x"]
                textString["VPOS"] = line["y"]
                textString["WIDTH"] = line["width"]
                textString["HEIGHT"] = line["height"]

                return textLine


            children = xml_page.findChildren(recursive=False)
            for child in children:
                if child.name == "textblock":

                    textBlock = soup_alto.new_tag("TextBlock")
                    printSpace.append(textBlock)

                    ex_blocks.append(textBlock)

                    textBlock["ID"] = "TextBlock_" + str(len(ex_blocks))

                    lines = child.find_all("line")

                    for line in lines:
                        textLine = createLine(line, ex_lines)
                        textBlock.append(textLine)
                        
                        

                elif child.name == "line":

                    textBlock = soup_alto.new_tag("TextBlock")
                    printSpace.append(textBlock)

                    ex_blocks.append(textBlock)

                    textBlock["ID"] = "TextBlock_" + str(len(ex_blocks))

                    textLine = createLine(child, ex_lines)
                    textBlock.append(textLine)

            output_path = os.path.join(output_dir, filename)
            os.makedirs(os.path.dirname(output_path), exist_ok=True)

            html = str(soup_alto)
            IIIF.format_and_save(html, output_path)
            IIIF.validate(output_path)

    @staticmethod
    def create_annolist(manifest3: dict, output_dir: str) -> None:
        """
        Create an annotation list.

        Args:
            manifest3: A dictionary containing the IIIF Presentation API 3.0 manifest data.
            output_dir: The directory where the annotation list JSON files should be saved.
        """
        items = manifest3["items"]

        for i, item in enumerate(items):
            filename = str(i + 1).zfill(4)
            output_path = output_dir + "/" + filename + ".json"

            resources = []

            for annotation in item["annotations"]:
                items = annotation["items"]

                for item in items:

                    target = item["target"]

                    if type (target) is str:
                        on = target
                    else:
                        on = [
                            {
                                "@type": "oa:SpecificResource",
                                "full": target["source"],
                                "selector": {
                                    "@type": "oa:Choice",
                                    "item": {
                                        "@type": "oa:SvgSelector",
                                        "value": target["selector"]["value"]
                                    }
                                }
                            }
                        ] # item["target"]

                    resource = {
                        "@id": item["id"],
                        "@type": "oa:Annotation",
                        "motivation": "sc:painting",
                        "resource": {
                            "@type": "cnt:ContentAsText",
                            "format": "text/plain",
                            "chars": item["body"]["value"]
                        },
                        "on": on
                    }
                    resources.append(resource)

            annolist = {
                "@context": "http://iiif.io/api/presentation/2/context.json",
                "@id": manifest3["id"].replace("/manifest.json", "") + "/canvases/" + filename + ".json",
                "@type": "sc:AnnotationList",
                "resources": resources
            }

            with open(output_path, "w") as f:
                json.dump(annolist, f, ensure_ascii=False, indent=4)


    @staticmethod
    def create_manifest_with_alto(manifest_path: str, output_path: str) -> None:
        """
        Create a manifest using ALTO format annotations.

        Args:
            manifest_path: The file path to the source IIIF Presentation API 3.0 manifest.
            output_path: The file path where the created IIIF Presentation API 2.0 manifest should be saved.
        """

        with open(manifest_path, "r") as f:
            manifest3 = json.load(f)

        IIIF.create_annolist(manifest3, os.path.dirname(manifest_path) + "/canvases")

        canvases2 = []
        canvases3 = manifest3["items"]

        prefix_url = manifest3["id"].replace("/manifest.json", "")

        for i, canvas3 in enumerate(canvases3):

            
            canvas2 = {
                "@id": canvas3["id"],
                "@type": "sc:Canvas",
                # "label": canvas3["label"],
                "height": canvas3["height"],
                "width": canvas3["width"],
                "images": [
                    {
                        "@type": "oa:Annotation",
                        "motivation": "sc:painting",
                        "on": canvas3["id"],
                        "resource": {
                            "@id": canvas3["items"][0]["items"][0]["body"]["id"],
                            "@type": "dctypes:Image",
                            "format": "image/jpeg",
                            "height": canvas3["height"],
                            "width": canvas3["width"],
                        }
                    }
                ],
                "otherContent": [
                    {
                        "@id": f"{prefix_url}/canvases/{str(i + 1).zfill(4)}.json",
                        "@type": "sc:AnnotationList"
                    }
                ],
                "seeAlso": {
                    "@id": f"{prefix_url}/canvases/{str(i + 1).zfill(4)}.xml",
                    "profile": "http://www.loc.gov/standards/alto/v4/alto.xsd",
                    "format": "text/xml",
                    "label": "METS-ALTO XML"
                }
            }

            canvases2.append(canvas2)

        manifest2 = {
            "@context": "http://iiif.io/api/presentation/2/context.json",
            "@id": manifest3["id"],
            "@type": "sc:Manifest",
            "label": manifest3["label"],
            # "metadata": manifest3["metadata"],
            "sequences": [
                {
                    # "@id": manifest3["sequences"][0]["@id"],
                    "@type": "sc:Sequence",
                    "canvases": canvases2
                }
            ]
        }

        with open(output_path, "w") as f:
            json.dump(manifest2, f, ensure_ascii=False, indent=4,
                    sort_keys=True, separators=(',', ': '))
            
    @staticmethod
    def ndl_ocr_post_process(root_path: str, target_path: str, output_iiif_v3_path: str, item_id: str, xml_path: str, 
                             output_anno_path: str, output_canvases_dir: str, opt_json_path: str, output_manifest_alto_path: str) -> None:
        """
        Process the NDL OCR and generate IIIF manifests and ALTO XMLs.

        Args:
            root_path (str): The root directory.
            target_path (str): Path to the target files.
            output_iiif_v3_path (str): Output path for IIIF v3 manifest.
            item_id (str): Identifier for the item.
            xml_path (str): Path to the input XML.
            output_anno_path (str): Output path for annotations.
            output_canvases_dir (str): Directory for saving the canvases.
            opt_json_path (str): Path to the options JSON.
            output_manifest_alto_path (str): Output path for manifest with ALTO annotations.
        """
        IIIF.create_iiif_manifest_from_local_dir(root_path, target_path, output_iiif_v3_path, item_id)
        IIIF.convert_ndl_ocr_xml_to_iiif_manifest(xml_path, output_iiif_v3_path, output_anno_path)
        IIIF.convert_ndl_ocr_xml_to_alto_xml(xml_path, output_canvases_dir, opt_json_path)
        IIIF.create_manifest_with_alto(output_anno_path, output_manifest_alto_path)

    @staticmethod
    def ndl_ocr_post_process_simple(root_path: str, output_dir: str, task_id: str) -> None:
        """
        Simplified NDL OCR post-processing. Generates IIIF manifests and ALTO XMLs based on a given task ID.

        Args:
            root_path (str): The root directory.
            output_dir (str): The output directory for the processed data.
            task_id (str): Identifier for the task.
        """
        target_path = f"{root_path}/files/{task_id}"
        base_manifest_path = f"{root_path}/iiif/{task_id}/manifest.json"
        xml_path = f"{output_dir}/{task_id}/xml/{task_id}.xml"
        output_manifest_path = f"{root_path}/iiif/{task_id}/manifest_a.json"
        output_canvases_dir = f"{root_path}/iiif/{task_id}/canvases"
        opt_json_path = f"{output_dir}/opt.json"
        output_manifest_alto_path = f"{root_path}/iiif/{task_id}/manifest_o.json"

        IIIF.ndl_ocr_post_process(root_path, target_path, base_manifest_path, task_id, xml_path, output_manifest_path, output_canvases_dir, opt_json_path, output_manifest_alto_path)

# %% ../nbs/iiif.ipynb 6
class Svg:
    """
    A class to represent a rectangle and its operations in the context of SVG manipulation.
    """

    def __init__(self):
        """
        Creates a new rectangle instance.
        """
        pass

    def extract_points_from_svg(self, svg_content):
        """
        Extract points from the SVG content.
        
        Parameters:
            svg_content (str): A string containing SVG data.

        Returns:
            np.array: A numpy array containing points extracted from the SVG.
        """
        doc = parseString(svg_content)
        path_strings = [path.getAttribute('d') for path in doc.getElementsByTagName('path')]
        points = []
        num_points_on_curve = 100
        for path_string in path_strings:
            path = parse_path(path_string)
            for e in path:
                if e.__class__.__name__ in ['Line', 'Move']:
                    points.append([e.start.real, e.start.imag])
                    points.append([e.end.real, e.end.imag])
                elif e.__class__.__name__ == 'CubicBezier':
                    t_values = np.linspace(0, 1, num_points_on_curve)
                    for t in t_values:
                        point = e.point(t)
                        points.append([point.real, point.imag])
                # Add more cases for other types of commands if needed
        return np.array(points, dtype=np.float32)

    def compute_minimum_area_rectangle(self, points):
        """
        Compute the minimum area rectangle for a given set of points.
        
        Parameters:
            points (np.array): A numpy array containing points to compute the rectangle for.

        Returns:
            tuple: A tuple containing center, angle, width and height of the rectangle.
        """
        rect = cv2.minAreaRect(points)
        center = rect[0]
        angle = rect[2]
        width = rect[1][0]
        height = rect[1][1]
        return center, angle, width, height

    def create_svg_with_rectangle(self, center, angle, width, height):
        """
        Create an SVG with a rectangle with given center, angle, width and height.
        
        Parameters:
            center (tuple): A tuple representing center of the rectangle.
            angle (float): Angle of rotation of the rectangle.
            width (float): Width of the rectangle.
            height (float): Height of the rectangle.

        Returns:
            svgwrite.drawing.Drawing: An SVG drawing with a rectangle.
        """
        dwg = svgwrite.Drawing(profile='tiny')
        
        angle_rad = -angle * np.pi / 180.0  # -angle to match SVG rotation direction
        rotation_matrix = np.array([
            [np.cos(angle_rad), -np.sin(angle_rad)],
            [np.sin(angle_rad), np.cos(angle_rad)]
        ])
        rect_points = np.array([
            [-width / 2, -height / 2],
            [width / 2, -height / 2],
            [width / 2, height / 2],
            [-width / 2, height / 2]
        ])
        rotated_points = (rotation_matrix @ rect_points.T).T + center
        path = dwg.path(d=f"M {rotated_points[0][0]} {rotated_points[0][1]} ", fill='none', stroke='black')
        for point in rotated_points[1:]:
            path.push(f"L {point[0]} {point[1]} ")
        path.push("Z")  # Close the path
        dwg.add(path)

        return dwg

    def convert_from_svg_str(self, svg):
        """
        Convert SVG string to an SVG string with a minimum area rectangle.
        
        Parameters:
            svg (str): An SVG string to convert.

        Returns:
            str: An SVG string with a minimum area rectangle.
        """
        points = self.extract_points_from_svg(svg)
        center, angle, width, height = self.compute_minimum_area_rectangle(points)
        svg_content = self.create_svg_with_rectangle(center, angle, width, height)
        return svg_content.tostring()

    def convert(self, path, opath):
        """
        Convert SVG content from JSON file to SVG content with minimum area rectangle.
        
        Parameters:
            path (str): The path to the input JSON file.
            opath (str): The path to the output JSON file.

        Returns:
            None
        """
        with open(path, 'r') as f:
            manifest_data = json.load(f)
        canvases = manifest_data['items']
        for canvas in canvases:
            annotations = canvas["annotations"][0]["items"]
            for annotation in annotations:
                target = annotation["target"]
                if "selector" in target and target["selector"]["type"] == "SvgSelector":
                    value = target["selector"]["value"]
                    points = self.extract_points_from_svg(value)
                    center, angle, width, height = self.compute_minimum_area_rectangle(points)
                    svg_content2 = self.create_svg_with_rectangle(center, angle, width, height)
                    target["selector"]["value"] = svg_content2.tostring()
        with open(opath, 'w') as f:
            json.dump(manifest_data, f, indent=2)
